/*
  ----------------------------------------------------------------------
  NOME: Roberto
  CONGNOME: Cavada
  MATRICOLA: xxx
  CODICE CALCOLATORE: 
  ----------------------------------------------------------------------
*/

#include <iostream>
#include <cstdlib>
#include <assert.h>

using namespace std;

/* --------------------------- INIZIO PUNTO 1  ------------------------- */
/*                    Scrivere qui sotto la soluzione                    */
/*    vvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvv   */

typedef unsigned int number;

// Singolo messaggio
struct Messaggio {
  char testo[100];
	number dest;
	number mitt;

	Messaggio() {
		testo[0] = '\0';
		dest = 0;
		mitt = 0;
	}

	Messaggio(const char* _testo, number _mitt, number _dest) {
		strcpy(testo, _testo);
		dest = _dest;
		mitt = _mitt;
	}
};


// Lista di messaggi
struct ListaMsg {
	Messaggio msg;
	ListaMsg* next;

	ListaMsg(const Messaggio &_msg, ListaMsg* _next=NULL) {
		msg = _msg;
		next = _next;
	}
};

// Coda di messaggi
struct CodaMsg {
	ListaMsg* lista;
	int len;

	CodaMsg() {
		lista = NULL;
		len = 0;
	}
};
    

// Server
struct Server {
	const char* nome;
	CodaMsg da_fibra; 
	CodaMsg a_fibra; 

	Server(const char* _nome) {
		nome = _nome;
	}
};



/*    ^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^   */
/* ---------------------------- FINE PUNTO 1  -------------------------- */



/* --------------------------- INIZIO PUNTO 2  ------------------------- */
/*                    Scrivere qui sotto la soluzione                    */
/*    vvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvv   */

/** Stampa il messaggio passato.
		se stampa_testo e' false, non stampa il testo del messaggio. 
		Formato richiesto:
Da <mitt> per <dest>: <testo>\n

    Se stampa_testo e' true, altimenti:
Da <mitt> per <dest>\n

   Ad esempio:
Da 123 a 321: Ciao!

   Notare il newline alla fine della riga. 
 */
void Messaggio_stampa(const Messaggio &msg, bool stampa_testo=true) 
{
	cout << "Da " << msg.mitt << " per " << msg.dest;
	if (stampa_testo) cout << ": " << msg.testo;
	cout << endl;
}


/** Stampa la lista di messaggi. Ogni messaggio su una riga diversa.
		Ad esempio:
Da 123 a 321: Ciao!
Da 321 a 123: Com'e' andato l'esame?
Da 123 a 321: Bene grazie!
 */
void ListaMsg_stampa(const ListaMsg* lista)
{
	for(const ListaMsg* iter = lista; iter != NULL; iter = iter->next) {
		Messaggio_stampa(iter->msg);
	}

}

/** Stampa la coda di messaggi, in qualsiasi ordine si desideri. 
    Il formato e':
Ci sono <len> messaggi in coda:
<msg1>
<msg2>
...
<msgN>

   Se stampa_messagi e' true e la coda non e' vuota, altrimenti:
Ci sono <len> messaggi in coda

   Si noti il newline alla fine. 
*/
void CodaMsg_stampa(const CodaMsg &coda, bool stampa_messaggi)
{
	cout << "Ci sono " << coda.len << " messaggi in coda";
	if (stampa_messaggi && coda.len > 0) {
		cout << ":\n";
		ListaMsg_stampa(coda.lista);
	}
	else cout << endl;
}

/** Stampa informazioni sul server. 
		Il formato e':
Server <nome>: 
Da fibra: <coda da_fibra>
A fibra: <coda a_fibra>

   Il parametro stampa_messaggi deve essere passato a CodaMsg_stampa. 
*/
void Server_stampa(const Server &server, bool stampa_messaggi)
{
	cout << "Server " << server.nome << ":\n";
	cout << "Da fibra: ";
	CodaMsg_stampa(server.da_fibra, stampa_messaggi);

	cout << "A fibra: ";
	CodaMsg_stampa(server.a_fibra, stampa_messaggi);
}


/*    ^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^   */
/* ---------------------------- FINE PUNTO 2 -------------------------- */



/* --------------------------- INIZIO PUNTO 3  ------------------------- */
/*                    Scrivere qui sotto la soluzione                    */
/*    vvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvv   */
void CodaMsg_push(CodaMsg &coda, const Messaggio &msg)
{
	coda.lista = new ListaMsg(msg, coda.lista);
	coda.len += 1;
}


Messaggio CodaMsg_pop(CodaMsg &coda)
{
	Messaggio res;

	// rimuovi ultimo elemento della lista:
	ListaMsg* iter = coda.lista;
	ListaMsg* prev = NULL;

	assert(coda.len > 0 && iter != NULL);

	while (true) {
		if (iter->next == NULL) {
			// rimuovi elemento puntato da iter
			if (prev != NULL) { // non e' l'elemento di testa
				prev->next = NULL;
			}
			else { // elemento di testa
				coda.lista = NULL;
			}
			
			res = iter->msg;
			delete iter;
			break;
		}

		prev = iter;
		iter = iter->next;
	}

	coda.len -= 1;
	return res;
}


/*    ^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^   */
/* ---------------------------- FINE PUNTO 3 -------------------------- */



/* --------------------------- INIZIO PUNTO 4  ------------------------- */
/*                    Scrivere qui sotto la soluzione                    */
/*    vvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvv   */


/** Riceve dall'antenna un messaggio */
void Server_ricevi_da_antenna(Server &server, const Messaggio &msg)
{
	CodaMsg_push(server.a_fibra, msg);
}


/** Rimuove un messaggio dalla coda da_fibra, e lo invia
		all'antenna. Stampa un testo */
void Server_invia_ad_antenna(Server &server)
{
	Messaggio msg = CodaMsg_pop(server.da_fibra);
	cout << server.nome << ": Inviato ad antenna messaggio: ";
	Messaggio_stampa(msg, false);
}

/** Riceve un messagio dalla fibra */
void Server_ricevi_da_fibra(Server &server, const Messaggio &msg)
{
	CodaMsg_push(server.da_fibra, msg);
}

/** Rimuove un messaggio dalla coda a_fibra, e restituisce il
		messaggio inviato. Stampa un testo. */
Messaggio Server_invia_a_fibra(Server &server)
{
	Messaggio msg = CodaMsg_pop(server.a_fibra);
	cout << server.nome << ": Inviato a fibra messaggio: ";
	Messaggio_stampa(msg, false);

	return msg;
}



/*    ^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^   */
/* ---------------------------- FINE PUNTO 4 -------------------------- */



// ----------------------------------------------------------------------
// Main del programma
// ----------------------------------------------------------------------

int main() 
{
	Server s1("Trento");
	Server s2("Verona");

	// 3 messaggi
	Messaggio messaggi[] = {
		Messaggio("Taca la pasta che 'rivo", 123, 321),
		Messaggio("Vaben, te 'speto!", 321, 123), 
		Messaggio("Ciao", 321, 213)
	};

	// 3 cellulari mandano un messaggio
	for (int i=0; i<(sizeof(messaggi)/sizeof(messaggi[0])); ++i) {
		Server_ricevi_da_antenna(s1, messaggi[i]);		
	}


	// Stampa server s1 e s2:
	Server_stampa(s1, true);
	cout << endl;
	Server_stampa(s2, true);

	// invia i messaggi sulla fibra, e trasferisci sul server s2
	for (int i=0; i<(sizeof(messaggi)/sizeof(messaggi[0])); ++i) {
		Messaggio msg = Server_invia_a_fibra(s1);
		Server_ricevi_da_fibra(s2, msg); // trasferisci a s2
	}

	// Stampa server s1 e s2:
	cout << endl;
	Server_stampa(s1, true);
	cout << endl;
	Server_stampa(s2, true);


	// trasferisci da s2 all'antenna
	for (int i=0; i<(sizeof(messaggi)/sizeof(messaggi[0])); ++i) {
		Server_invia_ad_antenna(s2);
	}
	

	// Stampa server s1 e s2:
	cout << endl;
	Server_stampa(s1, true);
	cout << endl;
	Server_stampa(s2, true);

  return 0;
}
